package com.huaching.parking.local.parking_local_general.service.local.base.chargeCalculator;

import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.TypeReference;
import com.huaching.parking.local.parking_local_general.domain.local.po.ThsChargingRulesPO;
import com.huaching.parking.local.parking_local_general.domain.local.pojo.chargeCalculator.BaseChargeDetailPOJO;
import com.huaching.parking.local.parking_local_general.domain.local.pojo.chargeCalculator.IntervalChargeDetailPOJO;
import com.huaching.parking.local.parking_local_general.enums.local.RequestChargeTypeEnums;
import com.huaching.parking.local.parking_local_general.util.local.DateUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.stereotype.Component;

import java.util.Calendar;
import java.util.Date;

/**
 * @program: parking_local_general
 * @description: 按照时间间隔计算费用
 * @author: lidongsheng
 * @createData: 2020-04-26 10:31
 * @updateAuthor: lidongsheng
 * @updateData: 2020-04-26 10:31
 * @updateContent: 按照时间间隔计算费用
 * @Version: 1.0.0
 * ************************************************
 * Copyright @ 李东升 2020. All rights reserved
 * ************************************************
 */


/*
//表示每2分钟收费10元
IntervalChargeDetail：
"data":{
    //金额
    "money": 100,
    // 0 分钟 1 小时
    "frameType": 0,
    // 时间
    "intervalTime": 2
}
*/

@Component
public class ThsIntervalPayableMoney implements ThsRequestChargeAbstract {

    private static final Logger logger = LoggerFactory.getLogger(ThsIntervalPayableMoney.class);

    @Override
    public Integer getPayableMoney(ThsChargingRulesPO thsChargingRulesPO, Date inTime,Date endTime) {
        String chargeDetails = thsChargingRulesPO.getChargeDetails();
        BaseChargeDetailPOJO<IntervalChargeDetailPOJO> baseChargeDetailPOJO =
                JSON.parseObject(chargeDetails, new TypeReference<BaseChargeDetailPOJO<IntervalChargeDetailPOJO>>() {
                });
        return intervalPayableMoney(baseChargeDetailPOJO, inTime, endTime, thsChargingRulesPO.getFreeTime()
                , thsChargingRulesPO.getMaxOneTimeMoney());
    }

    @Override
    public Integer getType() {
        return RequestChargeTypeEnums.INTERVAL.getCode();
    }

    /**
     * 按照间隔收费
     *
     * @param detail
     * @param inTime          计费开始时间
     * @param freeTime        免费时长（分钟）
     * @param maxOneTimeMoney 一次停车的最高费用（单位分）
     * @return
     */
    public Integer intervalPayableMoney(BaseChargeDetailPOJO<IntervalChargeDetailPOJO> detail, Date inTime, Date endTime,
                                        Integer freeTime, Integer maxOneTimeMoney) {
        inTime = DateUtils.getNextMinute(inTime, freeTime);
        //1. 判断是否在免费时长内
        if (inTime.after(endTime)) {
            return 0;
        }
        long sumMoney = detail.getFastMoney();
        //2. 判断是否在首时段内
        if (detail.validateFirstTime(detail, inTime, endTime)) {
            return Math.toIntExact(sumMoney);
        }
        //3. 加上首时段的时间
        int[] dateType = {Calendar.MINUTE, Calendar.HOUR_OF_DAY};
        inTime = DateUtils.calAddDate(inTime, detail.getFastTime(), dateType[detail.getFastFrameType()]);

        //4. 计算费用
        int dateNum = Math.toIntExact(DateUtils.subDate(inTime, endTime, detail.getData().getFrameType() + 2));
        sumMoney += (dateNum / detail.getData().getIntervalTime()) * detail.getData().getMoney();
        sumMoney = sumMoney > maxOneTimeMoney ? maxOneTimeMoney : sumMoney;
        return Math.toIntExact(sumMoney);

    }

    public static void main(String[] args) {
        ThsChargingRulesPO thsChargingRulesPO = new ThsChargingRulesPO();
        thsChargingRulesPO.setFreeTime(10);
        thsChargingRulesPO.setChargeType(RequestChargeTypeEnums.INTERVAL.getCode());
        thsChargingRulesPO.setMaxOneTimeMoney(20000000);
        thsChargingRulesPO.setChargeDetails("{\n" +
                "    \"fastTime\": 0,\n" +
                "    \"frameType\": 0,\n" +
                "    \"fastMoney\": 0,\n" +
                "    \"data\": {\n" +
                "        \"money\": 100,\n" +
                "        \"frameType\": 0,\n" +
                "        \"intervalTime\": 2\n" +
                "    }\n" +
                "}");
        ThsIntervalPayableMoney payableMoney = new ThsIntervalPayableMoney();
        Integer money = payableMoney.getPayableMoney(thsChargingRulesPO,
                DateUtils.stringConvertDate("2020-05-27 06:25:20", 3),
                DateUtils.stringConvertDate("2020-05-27 15:24:20", 3));
        System.out.println(money);
    }

}
