package com.huaching.parking.local.parking_local_general.service.local.base.chargeCalculator;

import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.TypeReference;
import com.huaching.parking.local.parking_local_general.domain.local.po.ThsChargingRulesPO;
import com.huaching.parking.local.parking_local_general.domain.local.pojo.chargeCalculator.BaseChargeDetailPOJO;
import com.huaching.parking.local.parking_local_general.domain.local.pojo.chargeCalculator.TimeFrameChargeDetailPOJO;
import com.huaching.parking.local.parking_local_general.enums.local.RequestChargeTypeEnums;
import com.huaching.parking.local.parking_local_general.util.local.DateUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.stereotype.Component;

import java.util.Calendar;
import java.util.Comparator;
import java.util.Date;
import java.util.List;

/**
 * @program: parking_local_general
 * @description: 按照时段收费
 * @author: lidongsheng
 * @createData: 2020-04-26 10:38
 * @updateAuthor: lidongsheng
 * @updateData: 2020-04-26 10:38
 * @updateContent: 按照时段收费
 * @Version: 1.0.0
 * ************************************************
 * Copyright @ 李东升 2020. All rights reserved
 * ************************************************
 */

/*
TimeFrameChargeDetail：
//json数组
        "data": [
            {
                //"frameType": 0 按分钟收费
                "frameType": 0,
                "beginTime": "00:00",
                "endTime": "07:00",
                //表示在00：00-07：00时间端内每分钟收费2元
                "money": 200
            },
            {
                //"frameType": 1 按小时收费
                "frameType": 1,
                "beginTime": "07:00",
                "endTime": "20:30",
                //表示在07:00-20:30时间端内每小时钟收费4元
                "money": 400
            },
            {
                //"frameType": 2 按固定收费
                "frameType": 2,
                "beginTime": "20:30",
                "endTime": "00:00",
                //表示在20:30-00:00时间端内一共收费3元
                "money": 300
            }
        ]

 */


@Component
public class ThsTimeFramePayableMoney implements ThsRequestChargeAbstract {

    private static final Logger logger = LoggerFactory.getLogger(ThsTimeFramePayableMoney.class);

    @Override
    public Integer getPayableMoney(ThsChargingRulesPO thsChargingRulesPO, Date inTime, Date endTime) {
        String chargeDetails = thsChargingRulesPO.getChargeDetails();
        BaseChargeDetailPOJO<List<TimeFrameChargeDetailPOJO>> detail =
                JSON.parseObject(chargeDetails, new TypeReference<BaseChargeDetailPOJO<List<TimeFrameChargeDetailPOJO>>>() {
                });
        return timeFramePayableMoney(detail, inTime, endTime, thsChargingRulesPO.getFreeTime()
                , thsChargingRulesPO.getMaxOneTimeMoney(), thsChargingRulesPO.getMaxOneDayMoney());
    }

    @Override
    public Integer getType() {
        return RequestChargeTypeEnums.TIME_FRAME.getCode();
    }

    /**
     * 按时段收费
     *
     * @param detail          按时段计费detail
     * @param inTime          进场时间
     * @param freeTime        免费时长（分钟）
     * @param maxOneTimeMoney 一次停车的最高费用（单位分）
     * @param maxOneDayMoney  一天停车的最高费用（单位分）
     * @return
     */
    public Integer timeFramePayableMoney(BaseChargeDetailPOJO<List<TimeFrameChargeDetailPOJO>> detail, Date inTime,
                                         Date endTime, Integer freeTime, Integer maxOneTimeMoney, Integer maxOneDayMoney) {
        inTime = DateUtils.getNextMinute(inTime, freeTime);
        //1. 判断是否在免费时长内
        if (inTime.after(endTime)) {
            return 0;
        }
        long sumMoney = detail.getFastMoney();
        //2. 判断是否在首时段内
        if (detail.validateFirstTime(detail, inTime, endTime)) {
            return Math.toIntExact(sumMoney);
        }
        //3. 加上首时段的时间
        int[] dateType = {Calendar.MINUTE, Calendar.HOUR_OF_DAY};
        inTime = DateUtils.calAddDate(inTime, detail.getFastTime(), dateType[detail.getFastFrameType()]);

        //4. 处理时段
        List<TimeFrameChargeDetailPOJO> list = detail.getData();
        list.sort(Comparator.comparing(o -> DateUtils.stringConvertDate("2020-01-01 " + o.getBeginTime(), 3)));
        list.stream().forEach(o -> {
            o.setBeginTime(o.getBeginTime() + ":00");
            o.setEndTime(o.getEndTime() + ":00");
        });
        //得到跨天的天数
        int acrossDayNum = (int) DateUtils.subDate(DateUtils.getNextDayStart(inTime, 0),
                DateUtils.getNextDayStart(endTime, 0), 4);
        //是否存在跨天
        if (acrossDayNum > 0) {
            logger.info("跨天费用计算");
            return acrossDayTimeFrameCharge(list, inTime, acrossDayNum, maxOneTimeMoney, maxOneDayMoney);
        } else {
            logger.info("非跨天费用计算");
            return noAcrossDayTimeFrameCharge(list, inTime, endTime, maxOneTimeMoney, maxOneDayMoney);
        }
    }

    /**
     * 按时段收费没有跨天的费用计算
     *
     * @param list
     * @param inTime          计费开始时间
     * @param endTime         计费结束时间
     * @param maxOneTimeMoney 一次停车的最高费用（单位分）
     * @param maxOneDayMoney  一天停车的最高费用（单位分）
     * @return
     */
    private Integer noAcrossDayTimeFrameCharge(List<TimeFrameChargeDetailPOJO> list, Date inTime, Date endTime
            , Integer maxOneTimeMoney, Integer maxOneDayMoney) {
        int sumMoney = 0;

        String tempInTime = DateUtils.dateFormatToString(inTime, 3) + " ";
        String tempEndTime = DateUtils.dateFormatToString(endTime, 3) + " ";
        for (TimeFrameChargeDetailPOJO detail : list) {
            Date detailBeginTime = DateUtils.stringConvertDate(tempInTime + detail.getBeginTime(), 3);
            Date detailEndTime = DateUtils.stringConvertDate(tempEndTime + detail.getEndTime(), 3);

            if (inTime.compareTo(detailBeginTime) >= 0
                    && endTime.compareTo(detailEndTime) <= 0) {
                sumMoney += frameTypeTimeFrameCharge(detail, inTime, endTime);
                break;
            }
            if (inTime.compareTo(detailBeginTime) >= 0
                    && inTime.compareTo(detailEndTime) < 0) {
                sumMoney += frameTypeTimeFrameCharge(detail, inTime, detailEndTime);
                continue;
            }

            if (inTime.compareTo(detailBeginTime) < 0
                    && endTime.compareTo(detailEndTime) >= 0) {
                sumMoney += frameTypeTimeFrameCharge(detail, null, null);
                continue;
            }
            if (endTime.compareTo(detailBeginTime) > 0
                    && endTime.compareTo(detailEndTime) < 0) {
                sumMoney += frameTypeTimeFrameCharge(detail, detailBeginTime, endTime);
                break;
            }

        }
        sumMoney = sumMoney > maxOneDayMoney ? maxOneDayMoney : sumMoney;
        sumMoney = sumMoney > maxOneTimeMoney ? maxOneTimeMoney : sumMoney;
        return sumMoney;
    }

    /**
     * 按时段收费跨天的费用计算
     *
     * @param list
     * @param inTime          开始时间（进场时间+freeTime）
     * @param acrossDayNum    跨天的天数
     * @param maxOneTimeMoney 一次停车的最高费用（单位分）
     * @param maxOneDayMoney  一天停车的最高费用（单位分）
     * @return
     */
    private Integer acrossDayTimeFrameCharge(List<TimeFrameChargeDetailPOJO> list, Date inTime, Integer acrossDayNum
            , Integer maxOneTimeMoney, Integer maxOneDayMoney) {

        int sumMoney = 0;
        //是否跨多天（跨天先计算出一整天需要缴费多少。）
        if (acrossDayNum > 1) {
            int oneDayMoney = oneDayTimeFrameCharge(list);
            logger.info("停车完整一天收费：" + oneDayMoney);
            oneDayMoney = oneDayMoney > maxOneDayMoney ? maxOneDayMoney : oneDayMoney;
            //所跨天数的费用
            int acrossDayMoney = oneDayMoney * (acrossDayNum - 1);
            sumMoney += acrossDayMoney;
            if (sumMoney > maxOneTimeMoney) {
                return maxOneTimeMoney;
            }
            logger.info("所跨天数停车的停车收费：" + sumMoney);
        }
        Date nowDate = new Date();
//        //记录最后一个区段的计费信息
//        TimeFrameChargeDetail lastDetail = list.get(list.size() - 1);

        long firstDayMoney = 0, endDayMoney = 0;

        //第一次循环是为了计算停车当天的总费用
        String tempInTime = DateUtils.dateFormatToString(inTime, 3) + " ";
        String tempNowTime = DateUtils.dateFormatToString(nowDate, 3) + " ";
        for (TimeFrameChargeDetailPOJO detail : list) {
            Date detailBeginTime = DateUtils.stringConvertDate(tempInTime + detail.getBeginTime(), 3);
            Date detailEndTime = DateUtils.stringConvertDate(tempInTime + detail.getEndTime(), 3);
            //最初开始区段
            if (inTime.compareTo(detailBeginTime) >= 0 &&
                    inTime.compareTo(detailEndTime) < 0
            ) {
                Date tempEnd = DateUtils.stringConvertDate(DateUtils.dateFormatToString(inTime, 3) + " " + detail.getEndTime(), 3);
                firstDayMoney += frameTypeTimeFrameCharge(detail, inTime, tempEnd);
            }
            //最初区段内
            if (inTime.compareTo(detailBeginTime) < 0
            ) {
                firstDayMoney += frameTypeTimeFrameCharge(detail, null, null);
            }
        }
        firstDayMoney = firstDayMoney > maxOneDayMoney ? maxOneDayMoney : firstDayMoney;
        logger.info("停车当天的停车收费：" + firstDayMoney);
        sumMoney += firstDayMoney;
        //第二次循环是为了计算停车结束当天的总费用
        for (TimeFrameChargeDetailPOJO detail : list) {
            Date detailNowBeginTime = DateUtils.stringConvertDate(tempNowTime + detail.getBeginTime(), 3);
            Date detailNowEndTime = DateUtils.stringConvertDate(tempNowTime + detail.getEndTime(), 3);
            //最终结束区段
            if (nowDate.compareTo(detailNowBeginTime) > 0 &&
                    nowDate.compareTo(detailNowEndTime) <= 0
            ) {
                Date tempBegin = DateUtils.stringConvertDate(tempNowTime + " " + detail.getBeginTime(), 3);
                endDayMoney += frameTypeTimeFrameCharge(detail, tempBegin, nowDate);
            }
            //结束区段内
            if (nowDate.compareTo(detailNowBeginTime) > 0
                    && nowDate.compareTo(detailNowEndTime) > 0
            ) {
                endDayMoney += frameTypeTimeFrameCharge(detail, null, null);
            }

        }
        endDayMoney = endDayMoney > maxOneDayMoney ? maxOneDayMoney : endDayMoney;
        logger.info("停车结束当天的停车收费：" + endDayMoney);
        sumMoney += endDayMoney;
        if (sumMoney > maxOneTimeMoney) {
            return maxOneTimeMoney;
        }
        return sumMoney;
    }

    /**
     * 停车一天的总费用
     *
     * @param list
     * @return
     */
    public Integer oneDayTimeFrameCharge(List<TimeFrameChargeDetailPOJO> list) {
        int oneDayMoney = 0;
        for (TimeFrameChargeDetailPOJO detail : list) {
            oneDayMoney += frameTypeTimeFrameCharge(detail, null, null);
        }
        return oneDayMoney;
    }

    /**
     * 按时段收费根据开始时间和结束时间的总费用
     *
     * @param detail
     * @param beginTime 如果有开始时间，就以给定的开始时间计算
     * @param endTime   如果有结束时间，就以给定的结束时间计算
     * @return
     */
    public long frameTypeTimeFrameCharge(TimeFrameChargeDetailPOJO detail, Date beginTime, Date endTime) {
        if (beginTime == null) {
            beginTime = DateUtils.stringConvertDate("2020-01-01 " + detail.getBeginTime(), 3);
        }
        if (endTime == null) {
            endTime = DateUtils.stringConvertDate("2020-01-01 " + detail.getEndTime(), 3);
        }
        switch (detail.getFrameType()) {
            //按照分钟收费
            case 0:
                long minute = DateUtils.subDate(beginTime, endTime, 2);
                return minute * detail.getMoney();
            //按照小时收费
            case 1:
                long hour = DateUtils.subDate(beginTime, endTime, 3);
                return hour * detail.getMoney();
            //按照固定收费
            case 2:
                return detail.getMoney();
        }
        return 0;
    }


    public static void main(String[] args) {
        ThsChargingRulesPO thsChargingRulesPO = new ThsChargingRulesPO();
        thsChargingRulesPO.setFreeTime(10);
        thsChargingRulesPO.setChargeType(RequestChargeTypeEnums.ONE_TIME.getCode());
        thsChargingRulesPO.setMaxOneTimeMoney(20000000);
        thsChargingRulesPO.setMaxOneDayMoney(20000000);
        thsChargingRulesPO.setChargeDetails("{\n" +
                "    \"fastTime\": 0,\n" +
                "    \"frameType\": 0,\n" +
                "    \"fastMoney\": 0,\n" +
                "    \"data\": [\n" +
                "        {\n" +
                "            \"frameType\": 0,\n" +
                "            \"beginTime\": \"00:00\",\n" +
                "            \"endTime\": \"07:00\",\n" +
                "            \"money\": 200\n" +
                "        },\n" +
                "        {\n" +
                "            \"frameType\": 1,\n" +
                "            \"beginTime\": \"07:00\",\n" +
                "            \"endTime\": \"20:30\",\n" +
                "            \"money\": 400\n" +
                "        },\n" +
                "        {\n" +
                "            \"frameType\": 2,\n" +
                "            \"beginTime\": \"20:30\",\n" +
                "            \"endTime\": \"00:00\",\n" +
                "            \"money\": 300\n" +
                "        }\n" +
                "    ]\n" +
                "}");
        ThsTimeFramePayableMoney payableMoney = new ThsTimeFramePayableMoney();
        Integer money = payableMoney.getPayableMoney(thsChargingRulesPO
                , DateUtils.stringConvertDate("2020-05-27 06:25:20", 3)
                , DateUtils.stringConvertDate("2020-05-27 15:26:28", 3));
        //期望结果：8600
        //计算结果：8600 正确
        System.out.println(money);
    }
}
